#!/usr/bin/env python

"""
visualize.py - Visualizes the effect of queries.

Copyright (C) 2010 David Boddie <david@boddie.org.uk>
Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
All rights reserved.
Contact: Nokia Corporation (qt-info@nokia.com)

This file is part of the documentation of Qt. It was originally
published as part of Qt Quarterly.

$QT_BEGIN_LICENSE:BSD$
You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
    the names of its contributors may be used to endorse or promote
    products derived from this software without specific prior written
    permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
$QT_END_LICENSE$
"""

import os, sys
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtXmlPatterns import *

class XmlItem:

    def __init__(self, name, row, parent = None):
    
        self.name = name
        
        # Record the item's location within its parent.
        self._row = row
        self.parentItem = parent
        self.childItems = []
        self.attributes = []
        self.value = TextItem(self)
        self.color = "white"

    def parent(self):
        return self.parentItem

    def child(self, i):
    
        attrs = len(self.attributes)
        if 0 <= i < attrs:
            return self.attributes[i]
        
        elif i < attrs + len(self.childItems):
            return self.childItems[i - attrs]
        
        return None
    
    def row(self):
        if self.parentItem:
            return len(self.parentItem.attributes) + self._row
        else:
            return self._row
    
    def setColor(self, color, parent_only = False):
    
        self.color = color
        
        if not parent_only:
            for item in self.attributes:
                item.setColor(color)
            for item in self.childItems:
                item.setColor(color)


class AttributeItem(XmlItem):

    def child(self, i):
        return None
    
    def row(self):
        return self._row


class TextItem:

    def __init__(self, parent):
        self.parentItem = parent
        self.value = u""
    
    def isEmpty(self):
        return self.value == u""
    
    def child(self, i):
        return None
    
    def row(self):
        return 0


class XmlItemModel(QAbstractItemModel):

    def __init__(self, root_item, parent = None):
    
        QAbstractItemModel.__init__(self, parent)
        
        self.rootItem = root_item
    
    def columnCount(self, parent):
        return 2
    
    def data(self, index, role):
    
        if not index.isValid():
            return QVariant()
        
        if role != Qt.DisplayRole and role != Qt.BackgroundRole:
            return QVariant()
        
        item = index.internalPointer()
        
        if role == Qt.BackgroundRole:
            return QVariant(QColor(item.color))
        
        if index.column() == 0:
            if isinstance(item, AttributeItem):
                return QVariant(item.name)
            else:
                return QVariant(u"<"+item.name+u">")
        
        elif index.column() == 1:
            if isinstance(item, AttributeItem):
                value = unicode(item.value)
            else:
                value = unicode(item.value.value)
            return QVariant(" ".join(value.split("\n")))
        else:
            return QVariant()
    
    def flags(self, index):
    
        if not index.isValid():
            return Qt.NoItemFlags
        
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role):
    
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:
                return QVariant(self.tr("Name"))
            elif section == 1:
                return QVariant(self.tr("Value"))
            else:
                return QVariant()
        
        return QVariant()
    
    def index(self, row, column, parent):
    
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()
        
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()
    
    def parent(self, child):
    
        if not child.isValid():
            return QModelIndex()
        
        childItem = child.internalPointer()
        parentItem = childItem.parent()
        
        if not parentItem or parentItem == self.rootItem:
            return QModelIndex()
        
        return self.createIndex(parentItem.row(), 0, parentItem)
    
    def rowCount(self, parent):
    
        if parent.column() > 0:
            return 0
        
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()
        
        if isinstance(parentItem, AttributeItem):
            return 0
        
        return len(parentItem.attributes) + len(parentItem.childItems)
    
    def setColor(self, color):
        self.rootItem.setColor(color)


class XmlItemModelBuilder(QAbstractXmlReceiver):

    def __init__(self, namePool):
    
        QAbstractXmlReceiver.__init__(self)
        self.namePool = namePool
        self.rootItem = None
        self._stack = []
    
    def attribute(self, xmlname, value):
    
        if self._stack:
            name = unicode(xmlname.localName(self.namePool))
            value = unicode(value.toString())
            item = self._stack[-1]
            attribute = AttributeItem(name, len(item.attributes), item)
            attribute.value = value
            item.attributes.append(attribute)
    
    def characters(self, value):
        if self._stack:
            self._stack[-1].value.value += unicode(value.toString())
    
    def comment(self, value):
        pass
    
    def namespaceBinding(self, xmlname):
        pass
    
    def startElement(self, xmlname):
        name = unicode(xmlname.localName(self.namePool))
        
        if self._stack:
            parentItem = self._stack[-1]
        else:
            self.rootItem = parentItem = XmlItem("#document", 0)
            self._stack.append(self.rootItem)
        
        row = len(parentItem.childItems)
        item = XmlItem(name, row, parentItem)
        
        parentItem.childItems.append(item)
        
        self._stack.append(item)
    
    def endElement(self):
        self._stack.pop()
    
    def startDocument(self):
        pass
    
    def endDocument(self):
        pass
    
    def startOfSequence(self):
        pass
    
    def endOfSequence(self):
        pass
    
    def itemModel(self):
        return XmlItemModel(self.rootItem)


class XmlNodeModel(QSimpleXmlNodeModel):

    def __init__(self, url, rootItem, namePool):
    
        QSimpleXmlNodeModel.__init__(self, namePool)
        
        self.url = url
        self.documentItem = rootItem
        self.cache = {}
    
    def documentUri(self, node_index):
    
        item = self._getItem(node_index)
        if item == self.documentItem:
            return self.url
        else:
            return QUrl()
    
    def _createIndex(self, item):
    
        # QXmlNodeModelIndex contains the item as data.
        self.cache[id(item)] = item
        return self.createIndex(id(item))
    
    def _getItem(self, node_index):
        pointer = node_index.data()
        if pointer:
            return self.cache.get(pointer)
        else:
            return None
    
    def root(self, node_index = None):
        return self._createIndex(self.documentItem)
    
    def name(self, node_index):
    
        item = self._getItem(node_index)
        if not item:
            return QXmlName()
        
        # Whenever we visit a node, darken the background color of the item,
        # making nodes that are visited frequently darker than others.
        item.setColor(QColor(item.color).darker(101), parent_only = True)
        return QXmlName(self.namePool(), item.name)
    
    def kind(self, node_index):
    
        item = self._getItem(node_index)
        
        # The caller guarantees that the node is not null, so we don't check
        # that here.
        
        if isinstance(item, AttributeItem):
            return QXmlNodeModelIndex.Attribute
        elif isinstance(item, TextItem):
            return QXmlNodeModelIndex.Text
        elif item.name == u"#document":
            return QXmlNodeModelIndex.Document
        else:
            return QXmlNodeModelIndex.Element
    
    def attributes(self, element):
    
        item = self._getItem(element)
        
        attributes = []
        for attr in item.attributes:
            attributes.append(self._createIndex(attr))
        return attributes
    
    def stringValue(self, node_index):
    
        item = self._getItem(node_index)
        if isinstance(item, TextItem):
            return item.value
        else:
            return QSimpleXmlNodeModel.stringValue(self, node_index)
    
    def typedValue(self, node_index):
    
        item = self._getItem(node_index)
        return QVariant(item.value)
    
    def compareOrder(self, ni1, ni2):
    
        if ni1 == ni2:
            return QXmlNodeModelIndex.Is
        else:
            item1 = self._getItem(ni1)
            item2 = self._getItem(ni2)
            if item1.row() < item2.row():
                return QXmlNodeModelIndex.Precedes
            elif item1.row() > item2.row():
                return QXmlNodeModelIndex.Follows
            else:
                return QXmlNodeModelIndex.Is
    
    def nextFromSimpleAxis(self, axis, origin):
    
        item = self._getItem(origin)
        if not item:
            return QXmlNodeModelIndex()
        
        if isinstance(item, AttributeItem):
        
            # An attribute
            if axis != self.Parent:
                return QXmlNodeModelIndex()
            
            # The parent of an attribute is the element which contains it.
            item = item.parentItem
        
        elif isinstance(item, TextItem):
        
            if axis == self.Parent:
                item = item.parentItem
            else:
                item = None
        
        else:
            # A regular element
            
            if axis == self.Parent:
                item = item.parentItem
            
            elif axis == self.FirstChild:
                # Navigate to the first element in the item's children.
                child = item.child(len(item.attributes))
                if child:
                    item = child
                elif not child and not item.value.isEmpty():
                    item = item.value
                else:
                    item = None
            
            elif axis == self.PreviousSibling:
                parent = item.parentItem
                # Only navigate as far back as the first element in the item's
                # collection of children.
                if parent and item.row() > len(parent.attributes):
                    item = parent.child(item.row() - 1)
                else:
                    item = None
            
            elif axis == self.NextSibling:
                parent = item.parentItem
                if parent:
                    item = parent.child(item.row() + 1)
                else:
                    item = None
        
        if item:
            return self._createIndex(item)
        else:
            return QXmlNodeModelIndex()
    
    def highlightNode(self, node_index):
    
        item = self._getItem(node_index)
        
        if isinstance(item, XmlItem):
            item.setColor("#ff9900")


class MessageHandler(QAbstractMessageHandler):

    def handleMessage(self, type, description, identifier, sourceLocation):
    
        self.errorBrowser.setHtml(description)


class MainWindow(QMainWindow):

    def __init__(self):
    
        QMainWindow.__init__(self)
        
        self.xmlPath = ""
        self.model = XmlItemModel(XmlItem(u"", 0))
        self.resultModel = XmlItemModel(XmlItem(u"", 0))
        self.messageHandler = MessageHandler()
        
        queries = [
            ('Whole document (books)', 'documents/books.xml', '$root'),
            ('All books', 'documents/books.xml', '$root//book'),
            ('All titles', 'documents/books.xml', '$root//title'),
            ('Qt for Symbian authors', 'documents/books.xml', '$root//title/text()[string() = "Qt for Symbian"]/../../authors'),
            ('Books with short titles', 'documents/books.xml', '$root//title/text()[string-length(string()) < 30]/../..'),
            ('Reformat document', 'documents/books.xml',
             '<books>\n'
             '{\n'
             '    for $book in $root//book\n'
             '        return <book title="{$book/title/text()}" authors="{$book/authors/text()}" />\n'
             '}\n'
             '</books>'),
            ('Whole document (forecast)', 'documents/forecast.xml', '$root'),
            ('Low temperatures', 'documents/forecast.xml', '$root//time/temperature[@value<10]'),
            ('Sun rise and temperatures', 'documents/forecast.xml',
             'for $time in $root//tabular/time\n'
             '     return <time>\n'
             '         { $time/@from }\n'
             '         <temperature>\n'
             '             { $time/temperature/@value, $time/temperature/@unit }\n'
             '         </temperature>\n'
             '         </time>'),
            ('Temperature and pressure', 'documents/forecast.xml',
             'for $time in $root//temperature[@value < 10]/../pressure[@value > 1030]/..\n'
             '  return <time>{ $time/@from, $time/@to } </time>'),
            ('Weather forecast', 'documents/forecast.xml',
             '$root//sun,\n'
             'for $time in $root//tabular/time\n'
             '    order by $time/@from\n'
             '    return <time>\n'
             '        { $time/@from }\n'
             '        <symbol>\n'
             '            { $time/symbol/@name, $time/symbol/@number }\n'
             '        </symbol>\n'
             '        <temperature>\n'
             '            { $time/temperature/@value, $time/temperature/@unit }\n'
             '        </temperature>\n'
             '        </time>')
            ]
        
        fileMenu = self.menuBar().addMenu(self.tr("&File"))
        fileMenu.addAction(self.tr("&Open..."), self.openFile, QKeySequence(self.tr("Ctrl+O")))
        fileMenu.addAction(self.tr("E&xit"), self.close, QKeySequence(self.tr("Ctrl+Q")))
        
        insertMenu = self.menuBar().addMenu(self.tr("&Insert"))
        for label, path, query in queries:
            action = QAction(label, self)
            action.setData(QVariant((path, query)))
            action.triggered.connect(self.insertQuery)
            insertMenu.addAction(action)
        
        queryMenu = self.menuBar().addMenu(self.tr("&Query"))
        queryMenu.addAction(self.tr("&Execute"), self.executeQuery, QKeySequence(self.tr("Ctrl+Return")))
        
        self.view = QTreeView()
        self.view.setModel(self.model)
        
        self.resultBrowser = QTextBrowser()
        self.resultView = QTreeView()
        self.resultView.setModel(self.resultModel)
        self.errorBrowser = QTextBrowser()
        self.messageHandler.errorBrowser = self.errorBrowser
        
        self.resultTabWidget = QTabWidget()
        self.resultTabWidget.addTab(self.resultBrowser, self.tr("&Output"))
        self.resultTabWidget.addTab(self.resultView, self.tr("&Structure"))
        self.errorTabIndex = self.resultTabWidget.addTab(self.errorBrowser, self.tr("&Errors"))
        
        self.queryEdit = QPlainTextEdit()
        self.queryEdit.setTabStopWidth(4)
        
        splitter = QSplitter(Qt.Vertical)
        splitter.addWidget(self.view)
        splitter.addWidget(self.resultTabWidget)
        splitter.addWidget(self.queryEdit)
        
        self.setCentralWidget(splitter)
        self.setWindowTitle(self.tr("Visualize XQuery"))
    
    def openFile(self):
    
        filePath = QFileDialog.getOpenFileName(self, self.tr("Open File"),
            self.xmlPath, self.tr("XML files (*.xml);;HTML files (*.html);;"
                                  "SVG files (*.svg);;User Interface files (*.ui);;"
                                  "All files (*)"))
        
        if not filePath.isEmpty():
            self._openFile(filePath)
    
    def _openFile(self, filePath):
    
        query = QXmlQuery()
        modelBuilder = XmlItemModelBuilder(query.namePool())
        query.bindVariable(u"url", QXmlItem(QVariant(filePath)))
        query.setQuery(u"declare variable $url external;\n"
                       u"doc($url)/*\n")
        query.evaluateTo(modelBuilder)
        
        self.model = modelBuilder.itemModel()
        self.view.setModel(self.model)
        self.view.expandAll()
        self.view.resizeColumnToContents(0)
        
        self.errorBrowser.clear()
        self.resetResults()
        self.queryEdit.setFocus(Qt.OtherFocusReason)
    
    def insertQuery(self):
    
        action = self.sender()
        path, query = action.data().toPyObject()
        dirPath = os.path.split(os.path.abspath(__file__))[0]
        self._openFile(os.path.join(dirPath, path))
        self.queryEdit.setPlainText(query)
    
    def executeQuery(self):
    
        # The query needs other components to use its name pool.
        query = QXmlQuery()
        query.setMessageHandler(self.messageHandler)
        nodeModel = XmlNodeModel(QUrl.fromLocalFile(self.xmlPath),
                                 self.model.rootItem, query.namePool())
        query.bindVariable(u"root", QXmlItem(nodeModel.root()))
        query_string = u"declare variable $root external;\n" + self.queryEdit.toPlainText()
        query.setQuery(query_string)
        
        self.model.setColor("white")
        
        if query.isValid():
        
            array = QByteArray()
            buf = QBuffer(array)
            buf.open(QIODevice.WriteOnly)
            
            formatter = QXmlFormatter(query, buf)
            
            if query.evaluateTo(formatter):
            
                self.resultBrowser.setPlainText(QString.fromUtf8(array))
                self.errorBrowser.clear()
            
            else:
                self.resultBrowser.clear()
                self.resultTabWidget.setCurrentIndex(self.errorTabIndex)
            
            buf.close()
            
            # Build the model that contains the data for the result view.
            
            modelBuilder = XmlItemModelBuilder(query.namePool())
            query.evaluateTo(modelBuilder)
            
            if modelBuilder.rootItem:
                self.resultModel = modelBuilder.itemModel()
                self.resultView.setModel(self.resultModel)
                self.resultView.expandAll()
                self.resultView.resizeColumnToContents(0)
                
                if self.resultTabWidget.currentIndex() == self.errorTabIndex:
                    self.resultTabWidget.setCurrentIndex(0)
            else:
                self.resultModel = XmlItemModel(XmlItem(u"", 0))
                self.resultView.setModel(self.resultModel)
            
            # Find the selected items in the original model.
            
            results = QXmlResultItems()
            query.evaluateTo(results)
            
            while True:
                item = results.next()
                if item.isNull():
                    break
                
                if item.isNode():
                    nodeModel.highlightNode(item.toNodeModelIndex())
            
            self.view.viewport().update()
        
        else:
            self.resetResults()
            self.resultTabWidget.setCurrentIndex(self.errorTabIndex)
        
    
    def resetResults(self):
    
        self.resultBrowser.clear()
        self.resultModel = XmlItemModel(XmlItem(u"", 0))
        self.resultView.setModel(self.resultModel)


if __name__ == "__main__":

    app = QApplication(sys.argv)
    window = MainWindow()
    window.resize(640, 512)
    window.show()
    
    sys.exit(app.exec_())
